package linearalgebra;

import linearalgebra.exception.InvalidSystemError;
import linearalgebra.exception.NoPivotError;
import linearalgebra.exception.OnlyZeroRowsError;
import linearalgebra.exception.SwapRowsException;
import linearalgebra.numeric.Matrix;
import linearalgebra.numeric.NumericColumn;
import linearalgebra.numeric.NumericColumnInterface;
import linearalgebra.numeric.NumericMatrixInterface;
import linearalgebra.numeric.NumericRowInterface;
import linearalgebra.numeric.TransposeWrapper;

public class LinearSystem {
	private NumericMatrixInterface L = null;
	private final NumericMatrixInterface a;
	private final UnknownValue[] x;
	private NumericColumnInterface b;
	private NumericColumnInterface c = null;
	private boolean m_partialPivot = true;
	
	private static final byte UNSOLVED = 0;
	private static final byte SINGULAR = 1;
	private static final byte INFINITE = 2;
	private static final byte ZEROSOLN = 3;
	private byte m_solvableStatus = UNSOLVED;
	
	public LinearSystem( NumericMatrixInterface m ) {
		this( m, true );
	}
	
	public LinearSystem( NumericMatrixInterface m, boolean partialPivot ) {
		this( m, new NumericColumn(m.getRowCount()), partialPivot );
	}
	
	public LinearSystem( NumericMatrixInterface m, NumericColumnInterface b ) {
		this.a = m;
		this.b = b;
		if( a.getRowCount() != b.getRowCount() ) {
			throw new InvalidSystemError();
		}
		this.x = new UnknownValue[a.getColumnCount()];
		for( int i = 0; i < this.x.length; i++ ) {
			this.x[i] = new UnknownValue(""+((char)('a'+i)));
		}
	}
	
	public LinearSystem( NumericMatrixInterface m, NumericColumnInterface b, boolean partialPivot ) {
		this( m, b );
		m_partialPivot = partialPivot;
	}
	
	public void solve() {
		PermutationMatrix p = new PermutationMatrix(a.getRowCount());
		if( L == null ) {
			L = (NumericMatrixInterface)a.factorL(p, m_partialPivot);
//			swapUnknowns(p);
			b.swapRows(p);
		}
		if( c == null ) {
			LinearSystem find_c = new LinearSystem( L, b );
			find_c.backSubstituteLower();
			c = find_c.getSolutionColumn();
		}
		backSubstituteUpper();
//		leftEliminate();
//		backSubstitute();
	}
	
	public NumericColumnInterface getSolutionColumn() {
		NumericColumn rc = new NumericColumn( b.getRowCount() );
		for( int i = 0; i < rc.getDimension(); i++ ) {
			rc.getNumericVectorView().setValue(i, getUnknownValue(i));
		}
		return rc;
	}
	
	public int getRowLeftPivotIndex( int row ) {
		return a.getRowLeftPivotIndex(row);
	}
	
	public double getUnknownValue( int row ) {
		return x[row].getNumericValue();
	}
	
	public void swapUnknowns( int row1, int row2 ) {
		UnknownValue temp = x[row1];
		x[row1] = x[row2];
		x[row2] = temp;
	}
	
	public UnknownValue getUnknown( int row ) {
		return x[row];
	}
	
	public void swapUnknowns( PermutationMatrix p ) {
		UnknownValue[] temp = new UnknownValue[x.length];
		for( int i = 0; i < x.length; i++ ) {
			temp[i] = x[ p.getRowTranspose(i) ];
		}
		System.arraycopy(temp, 0, x, 0, x.length);
	}
	
	public String getUnknownDescription( int index ) {
		return x[index].getDescription();
/*
		StringBuffer rc = new StringBuffer();
		NumericRowInterface row = a.getNumericRowView(index);
		double rightSum = rhs().getRowValue(index);
		double div = row.getColumnValue(index);
		rc.append( getVariableName(index) );
		rc.append( " = " );
		boolean bFirstRHS = true;
		for( int colIndex = 0; colIndex < row.getColumnCount(); colIndex++ ) {
			if( colIndex != index ) {
				double colValue = row.getColumnValue(colIndex);
				if( colValue != 0 ) {
					if( x[colIndex].isBound() ) {
						rightSum -= colValue * x[colIndex].getNumericValue();
					} else if( div != 0 ) {
						double printValue = -colValue/div;
						if( bFirstRHS ) {
							bFirstRHS = false;
						} else {
							if( printValue < 0 ) {
								rc.append( " - " );
							} else {
								rc.append( " + " );
							}
						}
						if( Math.abs(printValue) != 1.0 ) {
							rc.append( printValue );
							rc.append( "*" );
						}
						rc.append( getVariableName(colIndex) );
					}
				}
			}
		}
	
		if( div == 0.0 ) {
			if( rightSum == 0.0 ) {
				return getVariableName(index) + " is under constrained";
			} else {
				return getVariableName(index) + " is over constrained";
			}
		} else {
			if( rightSum !=  0 ) {
				if( !bFirstRHS ) {
					if( rightSum/div < 0 ) {
						rc.append( " - " );
					} else {
						rc.append( " + " );
					}
				}
				rc.append(Math.abs(rightSum/div));
			}
		}
		
		return rc.toString();
*/		
	}
	
	private String getVariableName( int index ) {
		return x[index].getName();
	}
	
	public void leftEliminate() {		
		try {
			for( int i=0; i < getEquationCount(); i++ ) {
				for( int j=i+1; j < getEquationCount(); j++ ) {
					try {
						double pivot = getPivot(i, j);
						combineRows(i, j, pivot);
					} catch( SwapRowsException e ) {
						swapRows( e.getRow1(), e.getRow2() );
						j--;
					}
				}
			}
		} catch( OnlyZeroRowsError e ) {
			// elimination is complete
		}
	}
	
	private void swapRows( int row1, int row2 ) {
		a.swapRows( row1, row2 );
		swapUnknowns( row1, row2 );
		rhs().swapRows( row1, row2 );
	}
	
	public void rightEliminate() {
		try {
			for( int i=getEquationCount()-1; i > 0; i-- ){
				for( int j=i-1; j >= 0; j-- ) {
					try {
						double pivot = getPivot(i, j);
						combineRows(i, j, pivot);
					} catch( SwapRowsException e ) {
						swapRows( e.getRow2(), e.getRow2() );
						j--;
					}
				}
			}
		} catch( OnlyZeroRowsError e ) {
			// elimination is complete
		}
	}
	
	public void normalize() throws SwapRowsException, OnlyZeroRowsError {
		for( int i=0; i < getEquationCount(); i++ ) {
			normalizeRow(i);
		}
	}
	
	public void backSubstituteUpper() {
		setSingular();	// initially assume the matrix is singular
		clearAllBindings(); // clear all unknown value bindings
		for( int rowIndex = a.getRowCount()-1; rowIndex >= 0; rowIndex-- ) {
			try {
				// try to get the pivot column for this row
				int unknownIndex = a.getRowLeftPivotIndex(rowIndex);
				// get the rhs of equation describing the pivot unknown variable
				x[unknownIndex].setConstraint(getUnknownConstraint(rowIndex, unknownIndex));
				// resolve previously bound variables
				x[unknownIndex].resolveBoundConstraints();
				// if there are still variable constraints, the system is underdetermined
				if( x[unknownIndex].getConstraint() != null ) {
					setUnderdeterminedUnknown();
				}
			} catch( NoPivotError e ) {
				// no pivot exists -- this is a zero row
				if( !Util.isZero(rhs().getRowValue(rowIndex)) ) {
					// The system is unsolvable.
					setInconsistentRow();
				}
			}
		}
	}
	
	public void backSubstituteLower() {
		setSingular();	// initially assume the matrix is singular
		clearAllBindings(); // clear all unknown value bindings
		for( int rowIndex = 0; rowIndex < a.getRowCount(); rowIndex++ ) {
			try {
				// try to get the pivot column for this row
				int unknownIndex = a.getRowRightPivotIndex(rowIndex);
				// get the rhs of equation describing the pivot unknown variable
				x[unknownIndex].setConstraint(getUnknownConstraint(rowIndex, unknownIndex));
				// resolve previously bound variables
				x[unknownIndex].resolveBoundConstraints();
				// if there are still variable constraints, the system is underdetermined
				if( x[unknownIndex].getConstraint() != null ) {
					setUnderdeterminedUnknown();
				}
			} catch( NoPivotError e ) {
				// no pivot exists -- this is a zero row
				if( !Util.isZero(rhs().getRowValue(rowIndex)) ) {
					// The system is unsolvable.
					setInconsistentRow();
				}
			}
		}
	}
/*	
	protected void bindRowUnknown( int unknownIndex, int rowIndex ) {
		int unknownCount = 0;
		int unknownZeroCount = 0;
		double lhs = 0.0;
		double div = 0.0;
		NumericRowInterface row = a.getNumericRowView(rowIndex);
		for( int acol_xrow = 0; acol_xrow < row.getColumnCount(); acol_xrow++ ) {
			double v = row.getColumnValue(acol_xrow);
			if( !Util.isZero(v) ) {
				if( !x[acol_xrow].isBound() ) {
					div = v;
					unknownCount++;
				} else {
					lhs += v * x[acol_xrow].getNumericValue();
				}
			} else {
				if( !x[acol_xrow].isBound() ) {
					unknownZeroCount++;
				}
			}
		}
		double total = rhs().getRowValue(rowIndex) - lhs;
		if( unknownCount > 0 ) {
			try {
				int unknownIndex = pivotIndex;
				if( unknownCount == 1 ) {
					double v = a.getElement(rowIndex, unknownIndex);
					x[unknownIndex].bindUnknown(total/v);
					if( m_solvableStatus == UNSOLVED ) {
						m_solvableStatus = SINGULAR;
					}
				} else {
					LinearExpression v = getUnknownConstraint(rowIndex, unknownIndex);
					x[unknownIndex].setConstraint(v);
					if( m_solvableStatus != ZEROSOLN ) {
						m_solvableStatus = INFINITE;
					}
				}
			} catch( NoPivotError e ) {
				throw new Error( "No pivot found on non-zero row." );
			}
		} else if( unknownZeroCount > 0 ) {
			// The entire row is zero.
			if( Util.isZero(total) ) {
				// this row allows any value - other rows may not
				if( m_solvableStatus != ZEROSOLN ) {
					m_solvableStatus = INFINITE;
				}
			} else {
				// no solution is possible
				m_solvableStatus = ZEROSOLN;
			}
		} else {
			if( !Util.isZero(total) ) {
				throw new Error( "All unknown values are bounds and system is inconsistent." );
			}
		}
		return;
	}
*/	
	public int getEquationCount() {
		return a.getRowCount();
	}
	
	public void multiplyRow( int row, double scalar ) {
		a.multiplyRow( row, scalar );
		rhs().multiplyRow( row, scalar );
	}
	
	public void normalizeRow( int row ) throws SwapRowsException, OnlyZeroRowsError {
		double n = 1.0 / a.getElement(row, getPivotColumn(row));
		multiplyRow(row, n);
	}
	
	public void combineRows( int srcRow, int dstRow, double scalar ) {
		a.combineRows(srcRow, dstRow, scalar);
		rhs().combineRows(srcRow, dstRow, scalar);
	}
	
	public String toString() {
		StringBuffer rc = new StringBuffer();
		for( int row = 0; row < a.getRowCount(); row++ ) {
			for( int col = 0; col < a.getColumnCount() ; col++ ) {
				rc.append( a.getElement(row, col) );
				if( col < a.getColumnCount() ) {
					rc.append("\t");
				}
			}
			rc.append("\t\t");
			if( row < x.length ) {
				rc.append(getVariableName(row));
			}
			rc.append("\t\t");
			rc.append(rhs().getRowValue(row));
			rc.append("\n");
		}
		for( int vars = a.getRowCount(); vars < a.getColumnCount(); vars++ ) {
			for( int col = 0; col < a.getColumnCount() ; col++ ) {
				rc.append("\t");
			}
			rc.append("\t");
			rc.append(getVariableName(vars));
			rc.append("\n");
		}
		
		return rc.toString();
	}
	
	public int getVariableCount() {
		return a.getColumnCount();
	}
	
	public double getPivot( int srcRow, int dstRow ) throws SwapRowsException, OnlyZeroRowsError {
		int pivotColumn = getPivotColumn(srcRow);
		double p1 = a.getElement(srcRow, pivotColumn);
		double p2 = a.getElement(dstRow, pivotColumn);
		return -p2/p1;
	}
	
	private int getPivotColumn( int row ) throws SwapRowsException, OnlyZeroRowsError {
		int rc = row;
		while( Util.isZero( a.getElement(row, rc) ) ) {
			// If all column values beneath this element are zero, move row to right
			for( int i = row+1; i < getEquationCount(); i++ ) {
				if( !Util.isZero(a.getElement(i, rc) ) ) {
					throw new SwapRowsException(row, i);
				}
			}
			rc++;
			if( rc >= getVariableCount() ) {
				throw new OnlyZeroRowsError();
			}
		}
		return rc;
	}
	
	public String getSolution( int row ) {
		return x[row].getDescription();
/*		
		StringBuffer rc = new StringBuffer();
		if( x[row].isBound() ) {
				rc.append( getVariableName(row) );
				rc.append( " = " );
				rc.append( x[row].getNumericValue() );
		} else {
			String constraint = getUnknownConstraint(row);
			if( constraint == null ) {
				rc.append(getVariableName(row));
				rc.append(" has no solution.");
			} else {
				rc.append( getVariableName(row) );
				rc.append( " = " );
				rc.append( constraint );
			}
		}
		return rc.toString();
*/		
	}
	
	private LinearExpression getUnknownConstraint( int row, int unknownColumn ) {
		LinearExpression rc = null;
		LinearEquation eqn = getSymbolicEquation(row);
		try {
			if( unknownColumn >= 0 && unknownColumn < x.length ) {
				eqn.solveFor(x[unknownColumn]);
				rc = eqn.getRHS();
			}
		} catch( NoPivotError e ) {
			// pivot was zero
		}
		return rc;
/* Implementation before LinearExpression - as String		
		double div = 0.0;
		int elementCount = 0;
		for( int col = 0; col < a.getColumnCount(); col++ ) {
			double v = a.getElement(row, col);
			if( col == unknownColumn ) {
				// we are at the variable we want
				if( Util.isZero(v) ) {
					// any solution - no constraint
					return null;
				}
				div = v;
			} else {
				v = -v;
				if( !Util.isZero(v) ) {
					v = appendSign( v, rc );
					if( v != 1.0 ) {
						rc.append(v);
						rc.append("*");
					}
					rc.append(getVariableName(col));
					elementCount++;
				}				
			}
		}
		double v = rhs().getElement(row, 0);
		if( !Util.isZero(v) ) {
			v = appendSign( v, rc );
			rc.append(v);
			elementCount++;
		}
		if( elementCount == 0 ) {
			rc.append("0.0");
			elementCount++;
		}
		if( elementCount > 1 && div != 1.0 ) {
			rc.insert(0, "( ");
			rc.append(" )");
		}
		if( div != 1.0 ) {
			rc.append("/");
			rc.append( div );
		}		
		return rc.toString();
*/		
	}
	
	private static double appendSign( double v, StringBuffer s ) {
		if( s.length() > 0 ) {
			if( v > 0 ) {
				s.append(" + ");
			} else {
				s.append(" - ");
			}		
			v = Math.abs(v);
		}
		return v;
	}
	
	private NumericColumnInterface rhs() {
		return c == null ? b : c;
	}
	
	public final boolean isNonsingular(){
		return m_solvableStatus == SINGULAR;
	}
	
	public final boolean isUnderdetermined() {
		return m_solvableStatus == INFINITE;
	}
	
	private final void setSingular() {
		m_solvableStatus = SINGULAR;
	}
	
	private final void setUnderdeterminedUnknown() {
		if( m_solvableStatus != ZEROSOLN ) {
			m_solvableStatus = INFINITE;
		}
	}
	
	private final void setInconsistentRow() {
		m_solvableStatus = ZEROSOLN;
	}
	
	private void clearAllBindings() {
		for( int i = 0; i < x.length; i++ ) {
			x[i].clearBindings();
		}
	}
	
	public final boolean isInconsistent() {
		return m_solvableStatus == ZEROSOLN;
	}
	
	public String getSolvableDescription() {
		String rc;
		switch( m_solvableStatus ) {
			case UNSOLVED: {
				rc = "System has not been solved.";
			} break;
			case SINGULAR: {
				rc = "System has a single solution.";
			} break;
			case INFINITE: {
				rc = "System has infinite solutions.";
			} break;
			case ZEROSOLN: {
				rc = "System has no solution.";
			} break;
			default: {
				throw new Error("System is in unknown state.");
			}
		}
		return rc;
	}
	
	public LinearEquation getSymbolicEquation( int rowIndex ) {
		LinearEquation rc = new LinearEquation();
		NumericRowInterface row = a.getNumericRowView( rowIndex );
		for( int i = 0; i < row.getColumnCount(); i++ ) {
			rc.addLeft( row.getColumnValue(i), x[i] );
		}
		rc.addRight( rhs().getRowValue(rowIndex) );
		return rc;
	}
	
	public void simplifyConstraints() {
		for( int i = 0; i < x.length; i++ ) {
			x[i].substituteConstraints();
		}
	}
	
	public GeneralSolution getGeneralSolution() {
		simplifyConstraints();
		int freeCount = 0;
		int[] freeIndex = new int[x.length];
		for( int i = 0; i < x.length; i++ ) {
			if( x[i].isFree() ) {
				freeIndex[freeCount++] = i;
			}
		}
		Matrix m = new Matrix( x.length, freeCount );
		for( int col = 0; col < m.getColumnCount(); col++ ) {
			UnknownValue w = x[freeIndex[col]];
			for( int unknownIndex = 0; unknownIndex < x.length; unknownIndex++ ) {
				if( x[unknownIndex] == w ) {
					m.setElement(unknownIndex, col, 1.0);
				} else if( x[unknownIndex].isBound() ) {
					// we will move this to the rhs?
				} else if( x[unknownIndex].isConstrained() ) {
					LinearExpression exp = x[unknownIndex].getConstraint();
					m.setElement(unknownIndex, col, exp.getCoefficient(w) );
				}
			}
		}
		double[] c = new double[m.getRowCount()];
		for( int unknownIndex = 0; unknownIndex < m.getRowCount(); unknownIndex++ ) {
			if( x[unknownIndex].isBound() ) {
				c[unknownIndex] = x[unknownIndex].getNumericValue();
			} else if( x[unknownIndex].isConstrained() ) {
				c[unknownIndex] = x[unknownIndex].getConstraint().getConstant();
			} else {
				c[unknownIndex] = 0.0;
			}
		}
		UnknownValue[] freeVariables = new UnknownValue[ freeCount ];
		for( int i = 0; i < freeCount; i++ ) {
			freeVariables[i] = new UnknownValue(x[freeIndex[i]].getName());
		}
		GeneralSolution rc = new GeneralSolution(m, freeVariables, new NumericColumn(c));
		return rc;
	}
	
	public int getRank() {
		solve();
		simplifyConstraints();
		int rc = 0;
		for( int i = 0; i < x.length; i++ ) {
			if( !x[i].isFree() ) {
				rc++;
			}
		}
		return rc;
	}
	
	public void setB( NumericColumn b ) {
		if( a.getRowCount() != b.getRowCount() ) {
			throw new InvalidSystemError();
		}
		this.b = b;
		this.c = null;
		clearAllBindings();
		m_solvableStatus = UNSOLVED;
	}
	
	public NumericColumnInterface getLeastSquaresSolution() {
		try {
			NumericMatrixInterface at = new TransposeWrapper(a);
			// TODO - this inverse should be a cholesky decomposition??
			NumericMatrixInterface atai = (NumericMatrixInterface)at.product(a).factorInverse();
			NumericMatrixInterface approx_x = atai.product(at).product(b);
			return approx_x.getNumericColumnView(0);
		} catch( linearalgebra.exception.UninvertibleMatrixException e ) {
			throw new Error("ata should be invertable.");
		}
	}
	
	public NumericColumnInterface getLeastSquaresProjection() {
		return a.product(getLeastSquaresSolution()).getNumericColumnView(0);
	}
	
	public double getLeastSquaresError() {
		return getLeastSquaresProjection().minus(b).getNumericColumnView(0).getNumericVectorView().getLengthSquared();
	}
	
	public NumericColumnInterface getLeastSquaresSolutionForOrthonomolColumns() {
		NumericMatrixInterface at = new TransposeWrapper(a);
		return at.product(b).getNumericColumnView(0);
	}
	
	private void shiftOriginToZero() {
		// need to choose a column on lhs and shift it to mean of zero
		// what to do with many columns
		// part of gram-schmitt
	}
}


