package linearalgebra.numeric;

import linearalgebra.*;
import linearalgebra.exception.*;

public abstract class AbstractNumericVector 
	implements NumericVectorInterface {
	
	public abstract double getValue( int dim );
	public abstract void setValue( int dim, double value );
	
	public Object clone() {
		try {
			return super.clone();
		} catch( CloneNotSupportedException e ) {
			throw new InternalError("CloneNotSupported unexpectedly caught in AbstractNumericVector.clone()");
		}
	}
	
	public double[] toArray() {
		double[] rc = new double[getDimension()];
		for( int i = 0; i < rc.length; i++ ) {
			rc[i] = getValue(i);
		}
		return rc;
	}
	
    public boolean equals(Object obj) {
    	if( !( obj instanceof NumericVectorInterface ) ) {
    		return false;
    	}
    	NumericVectorInterface v = (NumericVectorInterface)obj;
		if( getDimension() != v.getDimension() ) {
    		return false;
    	}
		for( int i = 0; i < getDimension(); i++ ) {
			if( getValue(i) != v.getValue(i) ) {
				return false;
			}
		}
        return true;
    }

    public double innerProduct(NumericVectorInterface v) {
		if( getDimension() != v.getDimension() ) {
			throw new SizeMismatchError("Vectors must be of same dimension to get inner product.");
		}
		double rc = 0.0;
		for( int i = 0; i < getDimension(); i++ ) {
			rc += getValue(i)*v.getValue(i);
		}
		return rc;
    }

    public void linearCombinationDimension(int src, int dst, double scalar) {
		setValue(dst, getValue(dst) + scalar*getValue(src));
    }
    
	public NumericVectorInterface linearCombination( NumericVectorInterface v, double scalar ) {
		NumericVector rc = (NumericVector)clone();
		rc.linearCombinationEquals(v, scalar);
		return rc;
	}

    public NumericVectorInterface linearCombinationEquals(NumericVectorInterface v, double scalar) {
		if( getDimension() != v.getDimension() ) {
			throw new SizeMismatchError("Vectors must be of same dimension to linearly combine.");
		}
		for( int i = 0; i < getDimension() ; i++ ) {
			setValue(i, getValue(i) + scalar*v.getValue(i) );
		}
		return this;
    }

    public NumericVectorInterface scale(double scalar) {
		for( int i = 0; i < getDimension(); i++ ) {
			scaleDimension(i, scalar);
		}
		return this;
    }

    public void scaleDimension(int dim, double scalar) {
		setValue(dim, scalar*getValue(dim) );
    }

    public void swapDimensions(int index1, int index2) {
		double temp = getValue(index1);
		setValue(index1, getValue(index2));
		setValue(index2, temp);
    }
    
    public final boolean isZero() {
    	boolean rc = true;
		for( int i = 0; i < getDimension(); i++ ) {
			if( !Util.isZero(getValue(i)) ) {
				rc = false;
				break;
			}
		}
		return rc;
    }

	public double getLengthSquared() {
		double rc = 0.0;
		for( int i = 0; i < getDimension(); i++ ) {
			rc += Math.pow( getValue(i), 2 );
		}
		return rc;
	}

	public double getLength() {
		return Math.sqrt(getLengthSquared());
	}

	public double getInnerProduct(VectorInterface v) {
		if( v.getDimension() != getDimension() ) {
			throw new SizeMismatchError("Array dimensions must match in getInnerProduct.");
		}
		double rc = 0.0;
		for( int i = 0; i < getDimension(); i++ ) {
			rc += v.getValue(i)*getValue(i);
		}
		return rc;
	}

	public boolean isLinearlyIndependent(VectorInterface v) {
		if( v.getDimension() != getDimension() ) {
			throw new SizeMismatchError("Array dimensions must match in getInnerProduct.");
		}
		boolean rc = true;
		if( !isZero() && !v.isZero() && v.getDimension() > 0 ) {
			double pivot = getValue(0)/v.getValue(0);
			for( int i = 1; i < getDimension(); i++ ) {
				if( pivot != getValue(i)/v.getValue(i) ) {
					rc = false;
					break;
				}
			}
		} else {
			rc = false;
		}
		/* also
		if( !isZero() && !v.isZero() && isOrthogonal() ) {
			return true;
		} else {
			// unknown
		}
		*/
		return rc;
	}
	
	public boolean isOrthogonal(VectorInterface v) {
		return Util.isZero(getInnerProduct(v));
	}
	
	/*
		Projects v onto this.
	*/
	public NumericVectorInterface projectionOf( NumericVectorInterface v ) {
		double d2 = getLengthSquared();
		double ip = getInnerProduct(v);
		scale(ip/d2);
		return this;
	}

	public NumericVectorInterface minus( NumericVectorInterface v ) {
		NumericVectorInterface rc = (NumericVectorInterface)clone();
		for( int i = 0; i < getDimension(); i++ ) {
			rc.setValue(i, rc.getValue(i) - v.getValue(i));
		}
		return rc;
	}
	
	public double getCosBetween( VectorInterface v ) {
		double innerProduct = getInnerProduct(v);
		double length1 = getLengthSquared();
		double length2 = v.getLengthSquared();
		return innerProduct/Math.sqrt(length1*length2);		
	}
	
	public double getRadiansBetween(VectorInterface v) {
		return Math.acos(getCosBetween(v));
	}


	public double getMean() {
		double sum = 0.0;
		for( int i = 0; i < getDimension(); i++ ) {
			sum += getValue(i);
		}
		return sum/getDimension();
	}
	
	public double getVariance() {
		double e = 0.0;
		for( int i = 0; i < getDimension(); i++ ) {
			e += Math.pow(getValue(i), 2);
		}
		return e;
	}
	
	public boolean normalize() {
		double l = getLength();
		if( Util.isZero(l) ) {
			return false;
		}
		scale(1.0/l);
		return true;
	}
}